/****************************************************************
    Project:                    Rapport Modular System
    This File:                  os_w32.c
    Last updated:               %U% %G%
    Version:                    %R%.%L%
    Last checked out:           %T% %H%
Copyright (c) 1993-1999 Rapport Limited. All rights reserved.
****************************************************************/


/*******************************
This version for:
  Machine:     IBM PC,
  Compiler:    Watcom,
  Environment: Win32 (95, NT, ...)
*******************************/


#include "os_w32.h"


/******************** RMS_File_Device_OS_Create ********************/

PIM_Error RMS_File_Device_OS_Create(RMS_File_Device_Pars *file_device)
{
  RMS_File_Device_Init(file_device);

  file_device->name = "os";

  file_device->device_data = (void*)NULL;

  file_device->dir_separator = '\\';
  file_device->ext_separator = '.';

  file_device->template_literal = 257;
  file_device->template_wild_one = '?';
  file_device->template_wild_many = '*';

  file_device->buffer_size = 64;
  file_device->buffer_replaceable = TRUE;

  file_device->device_destroy = RMS_File_Device_OS_Destroy;

  file_device->file_open = RMS_File_OS_Open;
  file_device->file_close = RMS_File_OS_Close;
  file_device->file_destroy = RMS_File_OS_Destroy;

  file_device->file_clearerror = RMS_File_OS_ClearError;

  file_device->file_startread = RMS_File_OS_StartRead;
  file_device->file_read = RMS_File_OS_Read;
  file_device->file_startwrite = RMS_File_OS_StartWrite;
  file_device->file_write = RMS_File_OS_Write;
  file_device->file_flush = RMS_File_OS_Flush;

  file_device->file_getposition = RMS_File_OS_GetPosition;
  file_device->file_setposition = RMS_File_OS_SetPosition;

  file_device->filename_absolute = RMS_File_OS_Absolute;
  file_device->filename_reformat = RMS_File_OS_Reformat;

  file_device->filename_delete = RMS_File_OS_Delete;
  file_device->filename_rename = RMS_File_OS_Rename;
  file_device->filename_copy = RMS_File_OS_Copy;

  file_device->filename_stat = RMS_File_OS_Stat;

  file_device->dir_open = RMS_File_OS_OpenDir;
  file_device->dir_close = RMS_File_OS_CloseDir;
  file_device->dir_read = RMS_File_OS_ReadDir;

  return(RMS_NO_ERROR);
}


/******************** RMS_File_Device_OS_Destroy ********************/

void RMS_File_Device_OS_Destroy(RMS_File_Device_Pars *file_device)
{
  RMS_File_Device_Init(file_device);
}


/******************** RMS_File_OS_Open ********************/

PIM_Error RMS_File_OS_Open(RMS_File_Pars *file_str,
                           char *filename,
                           PIM_Int access_flags)
{
  RMS_File_OS_Pars *file_os_pars;

  FILE *filepointer;


  /* init file struct */

  file_str->file_data = (void*)NULL;
  file_str->buffer_start = (PIM_Byte*)NULL;
  file_str->buffer_size = 0;


  /* try to open the file */

  switch (access_flags)
  {
    case RMS_FILE_ACCESS_READ :
    {
      filepointer = fopen(filename, "rb");

      break;
    }

    case RMS_FILE_ACCESS_WRITE | RMS_FILE_ACCESS_CREATE :
    {
      filepointer = fopen(filename, "wb");

      break;
    }

    case RMS_FILE_ACCESS_WRITE :
    case RMS_FILE_ACCESS_WRITE | RMS_FILE_ACCESS_APPEND :
    case RMS_FILE_ACCESS_READ | RMS_FILE_ACCESS_WRITE :
    case RMS_FILE_ACCESS_READ | RMS_FILE_ACCESS_WRITE | RMS_FILE_ACCESS_APPEND :
    {
      filepointer = fopen(filename, "r+b");

      break;
    }

    case RMS_FILE_ACCESS_READ | RMS_FILE_ACCESS_WRITE | RMS_FILE_ACCESS_CREATE:
    case RMS_FILE_ACCESS_READ | RMS_FILE_ACCESS_WRITE | RMS_FILE_ACCESS_CREATE | RMS_FILE_ACCESS_APPEND :
    {
      filepointer = fopen(filename, "w+b");

      break;
    }

    default :
    {
      return(RMS_ERROR_FILE_INVALIDACCESS);
    }
  }


  if (filepointer == (FILE*)NULL)
  {
    return(RMS_ERROR_UNDEFINED_FILENAME);
  }


  /* create an OS device file */

  file_os_pars = (RMS_File_OS_Pars*)
    file_str->file_device->memory_fns.malloc(file_str->file_device->memory_fns.memory_data,
                                             sizeof(RMS_File_OS_Pars));
  if (file_os_pars == (RMS_File_OS_Pars*)NULL)
  {
    fclose(filepointer);

    return(RMS_ERROR_OUT_OF_MEMORY);
  }


  /* store file parameters */

  file_os_pars->filepointer = filepointer;
  file_str->file_data = (void*)file_os_pars;


  return(RMS_NO_ERROR);
}


/******************** RMS_File_OS_Close ********************/

PIM_Error RMS_File_OS_Close(RMS_File_Pars *file_str)
{
  PIM_Error error_code = RMS_NO_ERROR;

  RMS_File_OS_Pars *file_os_pars = (RMS_File_OS_Pars*)file_str->file_data;


  if ((file_os_pars == (RMS_File_OS_Pars*)NULL) ||
      (file_os_pars->filepointer == (FILE*)NULL))
  {
    goto error_exit;
  }

  if (fclose(file_os_pars->filepointer))
  {
    error_code = RMS_ERROR_FILE_IO;
  }

error_exit:
  return(error_code);
}


/******************** RMS_File_OS_Destroy ********************/

void RMS_File_OS_Destroy(RMS_File_Pars *file_str)
{
  file_str->file_device->memory_fns.free(file_str->file_device->memory_fns.memory_data,
                                         (void*)file_str->buffer_start);
  file_str->buffer_start = (PIM_Byte*)NULL;

  file_str->file_device->memory_fns.free(file_str->file_device->memory_fns.memory_data,
                                         file_str->file_data);
  file_str->file_data = (void*)NULL;
}


/******************** RMS_File_OS_ClearError ********************/

void RMS_File_OS_ClearError(RMS_File_Pars *file_str)
{
  RMS_File_OS_Pars *file_os_pars = (RMS_File_OS_Pars*)file_str->file_data;


  clearerr(file_os_pars->filepointer);
}


/******************** RMS_File_OS_StartRead ********************/

PIM_Error RMS_File_OS_StartRead(RMS_File_Pars *file_str)
{
  /* check if need a buffer */

  if (file_str->buffer_start == (PIM_Byte*)NULL)
  {
    file_str->buffer_start = (PIM_Byte*)
      file_str->file_device->memory_fns.malloc(file_str->file_device->memory_fns.memory_data,
                                               file_str->file_device->buffer_size);
    if (file_str->buffer_start == (PIM_Byte*)NULL)
    {
      return(RMS_ERROR_OUT_OF_MEMORY);
    }

    file_str->buffer_size = file_str->file_device->buffer_size;

    file_str->last_access = RMS_FILE_ACCESS_READ;

    RMS_Error_Return(RMS_File_Buffer_Reset(file_str));
  }

  if (file_str->last_access != RMS_FILE_ACCESS_READ)
  {
    PIM_Int32 position;


    RMS_Error_Return(RMS_File_GetPosition(file_str,
                                          &position));

    file_str->last_access = RMS_FILE_ACCESS_READ;

    RMS_Error_Return(RMS_File_OS_SetPosition(file_str,
                                             position));
  }

  return(RMS_NO_ERROR);
}


/******************** RMS_File_OS_Read ********************/

PIM_Error RMS_File_OS_Read(RMS_File_Pars *file_str,
                           size_t *amount_read)
{
  RMS_File_OS_Pars *file_os_pars = (RMS_File_OS_Pars*)file_str->file_data;


  if (file_str->buffer_start == (PIM_Byte*)NULL)
  {
    RMS_Error_Return(RMS_File_OS_StartRead(file_str));
  }


  /* read into buffer */

  *amount_read = fread(file_str->buffer_start,
                       sizeof(char),
                       file_str->buffer_size,
                       file_os_pars->filepointer);
  if (*amount_read != 0)
  {
    return(RMS_NO_ERROR);
  }


  /* check for eof */

  if (feof(file_os_pars->filepointer))
  {
    return(RMS_ERROR_FILE_EOF);
  }

  return(RMS_ERROR_FILE_IO);
}


/******************** RMS_File_OS_StartWrite ********************/

PIM_Error RMS_File_OS_StartWrite(RMS_File_Pars *file_str)
{
  /* check if need a buffer */

  if (file_str->buffer_start == (PIM_Byte*)NULL)
  {
    file_str->buffer_start = (PIM_Byte*)
      file_str->file_device->memory_fns.malloc(file_str->file_device->memory_fns.memory_data,
                                               file_str->file_device->buffer_size);
    if (file_str->buffer_start == (PIM_Byte*)NULL)
    {
      return(RMS_ERROR_OUT_OF_MEMORY);
    }

    file_str->buffer_size = file_str->file_device->buffer_size;

    file_str->last_access = RMS_FILE_ACCESS_WRITE;

    RMS_Error_Return(RMS_File_Buffer_Reset(file_str));
  }

  if (file_str->last_access != RMS_FILE_ACCESS_WRITE)
  {
    PIM_Int32 position;


    RMS_Error_Return(RMS_File_GetPosition(file_str,
                                          &position));

    file_str->last_access = RMS_FILE_ACCESS_WRITE;

    RMS_Error_Return(RMS_File_OS_SetPosition(file_str,
                                             position));
  }

  return(RMS_NO_ERROR);
}


/******************** RMS_File_OS_Write ********************/

PIM_Error RMS_File_OS_Write(RMS_File_Pars *file_str,
                            size_t *amount_written)
{
  RMS_File_OS_Pars *file_os_pars = (RMS_File_OS_Pars*)file_str->file_data;

  size_t amount_to_write;


  if (file_str->buffer_start == (PIM_Byte*)NULL)
  {
    RMS_Error_Return(RMS_File_OS_StartWrite(file_str));
  }


  /* write buffer */

  amount_to_write = (size_t)(file_str->buffer_ptr - file_str->buffer_start);

  if (amount_to_write != 0)
  {
    if ((*amount_written = fwrite(file_str->buffer_start,
                                  sizeof(char),
                                  amount_to_write,
                                  file_os_pars->filepointer)) != amount_to_write)
    {
      return(RMS_ERROR_FILE_IO);
    }
  }

  return(RMS_NO_ERROR);
}


/******************** RMS_File_OS_Flush ********************/

PIM_Error RMS_File_OS_Flush(RMS_File_Pars *file_str)
{
  RMS_File_OS_Pars *file_os_pars = (RMS_File_OS_Pars*)file_str->file_data;


  /*** Decide what type of flush to do. ***/

  switch (file_str->last_access)
  {
    case RMS_FILE_ACCESS_READ :
    {
      if (fseek(file_os_pars->filepointer, 0, SEEK_END))
      {
        return(RMS_ERROR_FILE_IO);
      }

      break;
    }

    case RMS_FILE_ACCESS_WRITE :
    {
      RMS_Error_Return(RMS_File_Buffer_Output(file_str));

      if (fflush(file_os_pars->filepointer))
      {
        return(RMS_ERROR_FILE_IO);
      }

      break;
    }
  }


  return(RMS_File_Buffer_Reset(file_str));
}


/******************** RMS_File_OS_GetPosition ********************/

PIM_Error RMS_File_OS_GetPosition(RMS_File_Pars *file_str,
                                  PIM_Int32 *position)
{
  RMS_File_OS_Pars *file_os_pars = (RMS_File_OS_Pars*)file_str->file_data;


  *position = (PIM_Int32)ftell(file_os_pars->filepointer);

  if (*position == -1)
  {
    return(RMS_ERROR_FILE_IO);
  }

  return(RMS_NO_ERROR);
}


/******************** RMS_File_OS_SetPosition ********************/

PIM_Error RMS_File_OS_SetPosition(RMS_File_Pars *file_str,
                                  PIM_Int32 position)
{
  RMS_File_OS_Pars *file_os_pars = (RMS_File_OS_Pars*)file_str->file_data;


  if (fseek(file_os_pars->filepointer, (long)position, SEEK_SET))
  {
    return(RMS_ERROR_FILE_IO);
  }


  /* clear current buffer */

  return(RMS_File_Buffer_Reset(file_str));
}


/******************** RMS_File_OS_Delete ********************/

PIM_Error RMS_File_OS_Delete(RMS_File_Device_Pars *file_device,
                             char *filename)
{
  RMS_File_Pars file_str;

  PIM_Error error_code;


  error_code = RMS_File_OS_Stat(file_device,
                                filename,
                                &file_str);

  if ((!error_code) && (file_str.object_type == RMS_FILE_TYPE_FILE))
  {
    if (remove(filename))
    {
      return(RMS_ERROR_FILE_IO);
    }

    return(RMS_NO_ERROR);
  }

  return(RMS_ERROR_UNDEFINED_FILENAME);
}


/******************** RMS_File_OS_Rename ********************/

PIM_Error RMS_File_OS_Rename(RMS_File_Device_Pars *file_device,
                             char *old_filename,
                             char *new_filename)
{
  RMS_File_Pars file_str;

  PIM_Error error_code;


  error_code = RMS_File_OS_Stat(file_device,
                                old_filename,
                                &file_str);

  if ((!error_code) && (file_str.object_type == RMS_FILE_TYPE_FILE))
  {
    if (rename(old_filename, new_filename))
    {
      return(RMS_ERROR_FILE_IO);
    }

    return(RMS_NO_ERROR);
  }

  return(RMS_ERROR_UNDEFINED_FILENAME);
}


/******************** RMS_File_OS_Copy ********************/

PIM_Error RMS_File_OS_Copy(RMS_File_Device_Pars *file_device,
                           char *old_filename,
                           char *new_filename)
{
  RMS_File_Pars file_str;

  PIM_Error error_code;

  char temp_command[PIM_MAXIMUM_FILENAME_PATH_LENGTH * 2 + 16];


  error_code = RMS_File_OS_Stat(file_device,
                                old_filename,
                                &file_str);

  if ((!error_code) && (file_str.object_type == RMS_FILE_TYPE_FILE))
  {
    strcpy(temp_command, "copy ");
    strcat(temp_command, old_filename);
    strcat(temp_command, " ");
    strcat(temp_command, new_filename);

    if (system(temp_command))
    {
      return(RMS_ERROR_FILE_IO);
    }

    return(RMS_NO_ERROR);
  }

  return(RMS_ERROR_UNDEFINED_FILENAME);
}


/******************** RMS_File_OS_Stat ********************/

PIM_Error RMS_File_OS_Stat(RMS_File_Device_Pars *file_device,
                           char *filename,
                           RMS_File_Pars *file_str)
{
  struct stat stat_buffer;

  int error_code;


  /* check for first call to root dir */
/*
  if (*filename == 0)
  {
    file_str->object_type = RMS_FILE_TYPE_DIR;

    return(RMS_NO_ERROR);
  }
*/


  /* get real stat info */

  error_code = stat(filename, &stat_buffer);


  /* check for any error */

  if (error_code == -1)
  {
    return(RMS_ERROR_UNDEFINED_FILENAME);
  }


  /* object is there so extract the information */

  file_str->object_type = RMS_FILE_TYPE_UNKNOWN;

  if ((stat_buffer.st_mode & S_IFMT) == S_IFDIR)
  {
    file_str->object_type = RMS_FILE_TYPE_DIR;
  }
  else
  {
    if ((stat_buffer.st_mode & S_IFMT) == S_IFREG)
    {
      file_str->object_type = RMS_FILE_TYPE_FILE;
    }
  }

  file_str->object_size = (size_t)stat_buffer.st_size;

  file_str->created  = stat_buffer.st_ctime;
  file_str->accessed = stat_buffer.st_atime;
  file_str->modified = stat_buffer.st_mtime;

  return(RMS_NO_ERROR);
}


/******************** RMS_File_OS_OpenDir ********************/

void *RMS_File_OS_OpenDir(RMS_File_Device_Pars *file_device,
                          char *dirname)
{
  DIR *os_dfd;


  os_dfd = opendir(dirname);

  if (os_dfd == (DIR*)NULL)
  {
    char *dirname_end = dirname + strlen(dirname) - 1;

    if ((*dirname_end == file_device->dir_separator) ||
        (*dirname == 0))
    {
      strcat(dirname, ".");
    }
    else
    {
      strcat(dirname, "\\.");
    }

    os_dfd = opendir(dirname);

    dirname_end++;
    *dirname_end = 0;
  }

  return((void*)os_dfd);
}


/******************** RMS_File_OS_CloseDir ********************/

void RMS_File_OS_CloseDir(RMS_File_Device_Pars *file_device,
                          void *dir_dfd)
{
  DIR *os_dfd = (DIR*)dir_dfd;


  closedir(os_dfd);
}


/******************** RMS_File_OS_ReadDir ********************/

char *RMS_File_OS_ReadDir(RMS_File_Device_Pars *file_device,
                          void *dir_dfd)
{
  DIR *os_dfd = (DIR*)dir_dfd;
  struct dirent *os_dp;


re_readdir:
  os_dp = readdir(os_dfd);

  if (os_dp == (struct dirent*)NULL)
  {
    return((char*)NULL);
  }


  /* do not include . and .. */

  if ((!strcmp(os_dp->d_name, ".")) ||
      (!strcmp(os_dp->d_name, "..")))
  {
    goto re_readdir;
  }

  return((char*)os_dp->d_name);
}


/******************** RMS_File_OS_Absolute ********************/

PIM_Boolean RMS_File_OS_Absolute(RMS_File_Device_Pars *file_device,
                                 char *filename)
{
  if ((filename[0] == '\\') ||
      (filename[1] == ':'))
  {
    return(TRUE);
  }

  return(FALSE);
}


/******************** RMS_File_OS_Reformat ********************/

PIM_Int RMS_File_OS_Reformat(RMS_File_Device_Pars *file_device,
                             char *filename,
                             PIM_Int try_index)
{
  /*** 1) convert to native format ***/

  if (try_index == 1)
  {
    /*** Try to convert the file name to native format. ***/

    if (RMS_File_OS_Reformat_ToNative(filename))
    {
      return(0);
    }

    return(-1);
  }


#if FALSE
  /*** 2) EXAMPLE: prepend a home dir ***/

  if (try_index == 2)
  {
    RMS_File_OS_Pars *file_os_device_pars = (RMS_File_OS_Device_Pars*)device->device_data;
    char *home_dir;


    if (file_os_device_pars == (RMS_File_OS_Device_Pars*)NULL)
    {
      return(-1);
    }

    home_dir = file_os_device_pars->home_dir;

    if (home_dir == (char*)NULL)
    {
      return(-1);
    }


    if (!RMS_File_OS_Absolute(file_device,
                              filename))
    {
      PIM_Int path_length = strlen(home_dir);

      if (path_length)
      {
        char *new_filename = filename + path_length;

        if (home_dir[path_length - 1] == file_device->dir_separator)
        {
          memmove((void*)new_filename,
                  (void*)filename,
                  strlen(filename) + 1);
        }
        else
        {
          memmove((void*)&new_filename[1],
                  (void*)filename,
                  strlen(filename) + 1);

          *new_filename = file_device->dir_separator;
        }
        memcpy((void*)filename,
               (void*)home_dir,
                path_length);

        return(0);
      }
    }

    return(-1);
  }
#endif


  return(1);
}


/******************** RMS_File_OS_Reformat_ToNative ********************/

PIM_Boolean RMS_File_OS_Reformat_ToNative(char *filename)
{
  PIM_Int filename_format = RMS_FILE_OS_NATIVE;
  PIM_Boolean altered = FALSE;

  char *unix_check;


  /*** Check for a UNIX formatted filename. ***/

  unix_check = strchr(filename, '/');

  if (unix_check != (char*)NULL)
  {
    filename_format = RMS_FILE_OS_UNIX;
  }


  /*** Convert to native format. ***/

  switch (filename_format)
  {
    case RMS_FILE_OS_UNIX :
    {
      altered = RMS_File_OS_Reformat_UNIX(filename);
    }
  }


  return(altered);
}


/******************** RMS_File_OS_Reformat_UNIX ********************/

PIM_Boolean RMS_File_OS_Reformat_UNIX(char *filename)
{
  PIM_Boolean altered = FALSE;


  /*** Go through string. ***/

  while (*filename != 0)
  {
    if (*filename == '/')
    {
      *filename = '\\';

      altered = TRUE;
    }

    filename++;
  }


  return(altered);
}
