/****************************************************************
    Project:                    Platform Independent Module
    This File:                  pim.h
    Last updated:               %U% %G%
    Version:                    %R%.%L%
    Last checked out:           %T% %H%
Copyright (c) 1993-1999 Rapport Limited. All rights reserved.
****************************************************************/


/*** Avoid multiple includes ***/

#ifndef PIM_DOT_H
#define PIM_DOT_H


/*** base define's without the use of a prefix ***/

#ifndef TRUE
#define TRUE   1
#endif

#ifndef FALSE
#define FALSE  0
#endif

#if !TRUE
#error TRUE is defined to be FALSE
#endif

#if FALSE
#error FALSE is defined to be TRUE
#endif


/*** Hardware/Platform Parameter defines ***/

#define PIM_UNIX    1
#define PIM_DOS     2
#define PIM_WIN32   3
#define PIM_RISCOS  4
#define PIM_MACOS   5
#define PIM_LINUX   6

#define PIM_LOW_ORDER_FIRST   1
#define PIM_HIGH_ORDER_FIRST  2


/****************************************************************
Get the platform type
****************************************************************/

#ifndef PIM_PLATFORM

#ifdef unix
#ifdef __linux
#define PIM_PLATFORM   PIM_LINUX
#else
#define PIM_PLATFORM   PIM_UNIX
#endif
#endif


#ifdef __MWERKS__
#define PIM_PLATFORM   PIM_MACOS
#endif


#ifdef __riscos
#define PIM_PLATFORM   PIM_RISCOS
#endif


#ifdef __DOS__
#define PIM_PLATFORM   PIM_DOS
#endif


#ifdef __NT__
#define PIM_PLATFORM   PIM_WIN32
#endif

#ifdef _WIN32
#define PIM_PLATFORM   PIM_WIN32
#endif

#ifdef __WIN32__
#define PIM_PLATFORM   PIM_WIN32
#endif


#endif


/*** check platform is known ***/

#ifndef PIM_PLATFORM
#error unknown PIM_PLATFORM - check the compiler pre-defines above, or set it explicitly
#endif



/****************************************************************
Include the platform definition file
****************************************************************/

/* Linux machine */

#if PIM_PLATFORM == PIM_LINUX
#include "pim_linx.h"
#endif


/* Unix machine */

#if PIM_PLATFORM == PIM_UNIX
#include "pim_unix.h"
#endif


/* Mac machine */

#if PIM_PLATFORM == PIM_MACOS
#include "pim_mac.h"
#endif


/* RiscOS machine */

#if PIM_PLATFORM == PIM_RISCOS
#include "pim_risc.h"
#endif


/* DOS machine */

#if PIM_PLATFORM == PIM_DOS
#include "pim_dos.h"
#endif


/* WIN32 machine */

#if PIM_PLATFORM == PIM_WIN32
#include "pim_w32.h"
#endif


/******************** PIM Module ********************/

/*** define's ***/

#define PIM_MIN_INT32            -2147483648ul
#define PIM_MAX_INT32             2147483647
#define PIM_MIN_FLOAT_INT32      -(PIM_Float32)2147483648.0
#define PIM_SQRT_INT32            46340
#define PIM_SQRT_HALF_INT32       32768

#define PIM_FLT_EPSILON           1e-38
#define PIM_RELAXED_FLT_EPSILON   1e-6

#define PIM_NO_ERROR              0


/*** macro's ***/

#define PIM_EqualFloats(pim_op1, pim_op2)       (fabs((double)((pim_op1) - (pim_op2))) <= PIM_FLT_EPSILON)
#define PIM_ZeroFloat(pim_op1)                  (fabs((double)pim_op1) <= PIM_FLT_EPSILON)

#define PIM_NearEqualFloats(pim_op1, pim_op2)   (fabs((double)((pim_op1) - (pim_op2))) <= PIM_RELAXED_FLT_EPSILON)
#define PIM_NearZeroFloat(pim_op1)              (fabs((double)pim_op1) <= PIM_RELAXED_FLT_EPSILON)

#define PIM_SwapByteOrder2(pim_value) \
{ \
  PIM_Bytes2 pim_tmp; \
  PIM_Byte pim_byte; \
  pim_tmp.as_uint16 = (PIM_UInt16)pim_value; \
  pim_byte = pim_tmp.Byte_Order.byte0; \
  pim_tmp.Byte_Order.byte0 = pim_tmp.Byte_Order.byte1; \
  pim_tmp.Byte_Order.byte1 = pim_byte; \
  pim_value = pim_tmp.as_uint16; \
}

#define PIM_SwapByteOrder4(pim_value) \
{ \
  PIM_Bytes4 pim_tmp; \
  PIM_Byte pim_byte; \
  pim_tmp.as_uint32 = pim_value; \
  pim_byte = pim_tmp.Byte_Order.byte0; \
  pim_tmp.Byte_Order.byte0 = pim_tmp.Byte_Order.byte3; \
  pim_tmp.Byte_Order.byte3 = pim_byte; \
  pim_byte = pim_tmp.Byte_Order.byte1; \
  pim_tmp.Byte_Order.byte1 = pim_tmp.Byte_Order.byte2; \
  pim_tmp.Byte_Order.byte2 = pim_byte; \
  pim_value = pim_tmp.as_uint32; \
}


/*** typedef's ***/

typedef PIM_UInt8   PIM_Byte;
typedef PIM_Int     PIM_Boolean;
typedef PIM_Int     PIM_Error;

#if   PIM_BUS_BYTE_SIZE == 2
typedef PIM_Int16   PIM_IntBus;

#elif PIM_BUS_BYTE_SIZE == 4
typedef PIM_Int32   PIM_IntBus;

#elif PIM_BUS_BYTE_SIZE == 8
typedef PIM_Int64   PIM_IntBus;

#else
#error un-supported PIM_BUS_BYTE_SIZE

#endif

#if PIM_BYTE_ORDER == PIM_HIGH_ORDER_FIRST
typedef union
{
  struct
  {
    PIM_Byte  byte0;
    PIM_Byte  byte1;
    PIM_Byte  byte2;
    PIM_Byte  byte3;
  } Byte_Order;

  struct
  {
    PIM_Byte  byte3;
    PIM_Byte  byte2;
    PIM_Byte  byte1;
    PIM_Byte  byte0;
  } Byte_Value;

  PIM_Int32    as_int32;

  PIM_UInt32   as_uint32;

  PIM_Float32  as_float32;

} PIM_Bytes4;

typedef union
{
  struct
  {
    PIM_Byte  byte0;
    PIM_Byte  byte1;
  } Byte_Order;

  struct
  {
    PIM_Byte  byte1;
    PIM_Byte  byte0;
  } Byte_Value;

  PIM_Int16   as_int16;

  PIM_UInt16  as_uint16;

} PIM_Bytes2;

#else
typedef union
{
  struct
  {
    PIM_Byte  byte0;
    PIM_Byte  byte1;
    PIM_Byte  byte2;
    PIM_Byte  byte3;
  } Byte_Order;

  struct
  {
    PIM_Byte  byte0;
    PIM_Byte  byte1;
    PIM_Byte  byte2;
    PIM_Byte  byte3;
  } Byte_Value;

  PIM_Int32    as_int32;

  PIM_UInt32   as_uint32;

  PIM_Float32  as_float32;

} PIM_Bytes4;

typedef union
{
  struct
  {
    PIM_Byte  byte0;
    PIM_Byte  byte1;
  } Byte_Order;

  struct
  {
    PIM_Byte  byte0;
    PIM_Byte  byte1;
  } Byte_Value;

  PIM_Int16   as_int16;  

  PIM_UInt16  as_uint16;

} PIM_Bytes2;

#endif


/*** prototype's ***/


#endif
